

// Maze generation class for roguelike maze generator
// Copyright Gerry Quinn
// 25th August, 2007


// Uncomment if using pre-compiled header on MSVC
// #include "stdafx.h"


#include "Maze.h"


#include <vector>
using std::swap;


#include <algorithm>
using std::random_shuffle;


#include <functional>
using std::pointer_to_unary_function;


// The only thing used from the DungeonGenerator classes, and I just simulate MSVC rand()
#include "MersenneTwister.h"



//////////////////////////////////////////////////////////////////////
// Static


MTRand Maze::m_Rand;


// Simulates MSVC rand() but uses Mersenne Twister
int Maze::random()
{
	unsigned int val = (unsigned int)( m_Rand.randInt() );
	return (int)( val & 0x7fff );
}


// Used in algorithm::random_shuffle()
int Maze::rsRand( int n )
{
	return( random() % n );
}


//////////////////////////////////////////////////////////////////////
// MazeParams


Maze::MazeParams::MazeParams()
{
	m_Size = Point( 80, 48 );
	m_CorridorWidth = 2;
	m_WallWidth = 2;
	m_nRooms = 7;
	m_RoomMin = Point( 8, 8 );
	m_RoomMax = Point( 12, 12 );
	m_nLoops = 10;
	m_nDEPasses = 2;
}


void Maze::MazeParams::GetUnexpandedRoomSizes( Point & minDim, Point & maxDim ) const
{
	int exp = GetExpansionFactor();
	minDim = Point( ( m_RoomMin.x + 1 ) / exp, ( m_RoomMin.y + 1 ) / exp );
	maxDim = Point( ( m_RoomMax.x + 1 ) / exp, ( m_RoomMax.y + 1 ) / exp );
}

void Maze::MazeParams::GetActualRoomSizes( Point & minRoom, Point & maxRoom ) const
{
	minRoom = m_RoomMin;
	maxRoom = m_RoomMax;
}

int Maze::MazeParams::GetDeadEndPasses() const
{
	return m_nDEPasses;
}

Point Maze::MazeParams::GetSize() const
{
	return m_Size;
}

int Maze::MazeParams::GetCorridorWidth() const
{
	return m_CorridorWidth;
}

int Maze::MazeParams::GetWallWidth() const
{
	return m_WallWidth;
}

int Maze::MazeParams::GetNumberOfRooms() const
{
	return m_nRooms;
}

int Maze::MazeParams::GetNumberOfLoops() const
{
	return m_nLoops;
}

int Maze::MazeParams::GetExpansionFactor() const 
{ 
	return m_CorridorWidth + m_WallWidth;
}


//////////////////////////////////////////////////////////////////////////
// Construction/Destruction/Initialisaton
//////////////////////////////////////////////////////////////////////


Maze::Maze()
{
	m_Rand.seed();

	m_DirFlags[ DIR_N ] = CF_N;
	m_DirFlags[ DIR_E ] = CF_E;
	m_DirFlags[ DIR_S ] = CF_S;
	m_DirFlags[ DIR_W ] = CF_W;

	m_Deltas[ 0 ] = Point(  0,  1 );
	m_Deltas[ 1 ] = Point(  1,  0 );
	m_Deltas[ 2 ] = Point(  0, -1 );
	m_Deltas[ 3 ] = Point( -1,  0 );
}

Maze::~Maze()
{

}

// Set the parameters based on requested parameters
// WARNING! There is some sanity checking, but no guarantees!
// Do not even think about inserting negative numbers...
// Note: actual room dimensions will always be one less than a multiple of (corridor width + wall width) 
void Maze::SetParams( const MazeParams & requestedParams )
{
	MazeParams mp = requestedParams;
	
	if ( mp.m_Size.x > 200 ) mp.m_Size.x = 200;
	if ( mp.m_Size.y > 200 ) mp.m_Size.y = 200;
	if ( mp.m_CorridorWidth == 0 ) mp.m_CorridorWidth = 1;
	if ( mp.m_WallWidth == 0 ) mp.m_WallWidth = 1;
	while( mp.GetExpansionFactor() > EXP_MAXMUL )
	{
		mp.m_CorridorWidth--;
		mp.m_WallWidth--;
		if ( mp.m_CorridorWidth == 0 ) mp.m_CorridorWidth = 1;
		if ( mp.m_WallWidth == 0 ) mp.m_WallWidth = 1;
	}
	while ( mp.m_RoomMin.x < mp.GetExpansionFactor() * 2 - 1 ) mp.m_RoomMin.x++;
	while ( mp.m_RoomMin.y < mp.GetExpansionFactor() * 2 - 1 ) mp.m_RoomMin.y++;
	if ( mp.m_RoomMax.x < mp.m_RoomMin.x ) mp.m_RoomMax.x = mp.m_RoomMin.x;
	if ( mp.m_RoomMax.y < mp.m_RoomMin.y ) mp.m_RoomMax.y = mp.m_RoomMin.y;

	m_Params = mp;
}


////////////////////////////////////////////////////////////////////////


// Generate a maze
void Maze::Create()
{
	const int exp = m_Params.GetExpansionFactor();				

	const int nPassesDE = m_Params.GetDeadEndPasses();

	// Set the size of the unexpanded board
	{
		Point fullSize = m_Params.GetSize();
		int wall = m_Params.GetWallWidth();
		m_Size.x = ( fullSize.x - wall ) / exp;
		m_Size.y = ( fullSize.y - wall ) / exp;
	}

	// Initialise a big empty maze
	{
		for ( int x = 1; x <= m_Size.x; x++ )
		{
			for ( int y = 1; y <= m_Size.y; y++ )
			{
				Point pos( x, y );
				m_Board[ pos.x ][ pos.y ] = m_DirFlags[ DIR_N ] + m_DirFlags[ DIR_E ] + m_DirFlags[ DIR_S ] + m_DirFlags[ DIR_W ];
				if ( x == 1 ) MakeWall( pos, DIR_W, true );
				if ( x == m_Size.x ) MakeWall( pos, DIR_E, true );
				if ( y == 1 ) MakeWall( pos, DIR_S, true );
				if ( y == m_Size.y ) MakeWall( pos, DIR_N, true );
			}
		}
	}

	// Add some rooms, which cannot be changed
	AddRooms();

	// Fill the rest of the map with a connected, tree-structured maze
	FillWithMaze();

	// Break some walls to get loops
	AddLoops();

	// Remove some dead-ends
	RemoveDeadEnds( nPassesDE );

	// Expand it
	Expand();

	// Done!
}



// Use Dijkstra method to expand from a starting square in m_Wave
// Starting point is set to 1, adjacent points to 2, etc.  Unreached points are set to 0.
// Returns number of squares found (including starting square)
// Can be used for pathfinding, connectivity check, etc.
// Mode MWM_COUNT - fills all, returns number found
// Mode MWM_DEADEND - fills from dead-end to fork, returns number in fork square, returns square preceding fork square
int Maze::MakeWave( const Point & startPos, MWMode mode, Point & deadendFork )
{
	for ( int x = 1; x <= m_Size.x; x++ )
	{
		for ( int y = 1; y <= m_Size.y; y++ )
		{
			m_Wave[ x ][ y ] = 0;
		}
	}
	
	vector< Point > proc;			// squares we are processing
	vector< Point > store;			// new squares we found adjacent to those in proc
	int count = 0;					// record total number of squares reached

	int dist = 1;
	m_Wave[ startPos.x ][ startPos.y ] = dist;
	store.push_back( startPos );
	count++;

	while ( true )
	{
		swap( proc, store );
		store.clear();
		dist++;
		for ( int iProc = 0; iProc < proc.size(); iProc++ )
		{
			const Point & pos = proc[ iProc ];
			for ( int dir = 0; dir < 4; dir++ )
			{
				if ( ! IsWall( pos, dir ) )
				{
					Point adj = GetAdj( pos, dir );
					if ( m_Wave[ adj.x ][ adj.y ] == 0 )
					{
						m_Wave[ adj.x ][ adj.y ] = dist;
						store.push_back( adj );
						count++;
					}
				}
			}			
		}

		switch( mode )
		{
			case MWM_COUNT:
				{
					if ( store.empty() ) 
					{
						return count;
					}
				}
				break;

			case MWM_DEADEND:
				{
					if ( store.size() > 1 )
					{
						deadendFork = proc[ 0 ];
						return m_Wave[ deadendFork.x ][ deadendFork.y ];
					}
				}
				break;

			default:
				return -1;
		}

		
		if ( store.empty() ) break;					// shouldn't ever happen
	}

	return count;
}

// Find a path from startPos to start point of wavefront algorithm (wavefront assumed to have been done)
// Path is inclusive of start and end points
void Maze::FindPath( const Point & startPos, vector< Point > & path ) const
{
	path.clear();
	Point pos = startPos;
	int val = m_Wave[ pos.x ][ pos.y ];
	while( true )
	{
		path.push_back( pos );
		val -= 1;
		if ( val == 0 ) return;
		int iDir;
		for ( iDir = 0; iDir < 4; iDir++ )
		{
			Point adj = pos + m_Deltas[ iDir ];
			if ( m_Wave[ adj.x ][ adj.y ] == val )
			{
				pos = adj;
				break;
			}
		}
		if ( iDir == 4 )
		{
			// Error!
			path.clear();
			return;
		}
	}
}


// Make a wall in this direction from this cell (clear it if make is false)
void Maze::MakeWall( const Point & pos, int dir, bool make )
{
	Point adj = GetAdj( pos, dir );
	if ( make )
	{
		m_Board[ pos.x ][ pos.y ] &= ~ m_DirFlags[ dir ];
		m_Board[ adj.x ][ adj.y ] &= ~ m_DirFlags[ ( dir + 2 ) & 3 ]; 
	}
	else
	{
		m_Board[ pos.x ][ pos.y ] |= m_DirFlags[ dir ];
		m_Board[ adj.x ][ adj.y ] |= m_DirFlags[ ( dir + 2 ) & 3 ]; 
	}
}


// Expand the maze to a walled region 
void Maze::Expand()
{
	const int exp = m_Params.GetExpansionFactor();

	const int wall = m_Params.GetWallWidth();

	// Completely fill with walls
	{
		for ( int x = 0; x < MAZE_XMAX; x++ )
		{
			for ( int y = 0; y < MAZE_YMAX; y++ )
			{
				m_Exp[ x ][ y ] = T_WALL;
			}
		}
	}

	// Expand the maze
	{
		for ( int x = 1; x <= m_Size.x; x++ )		
		{
			for ( int y = 1; y <= m_Size.y; y++ )
			{
				int xP = wall + ( x - 1 ) * exp;
				int yP = wall + ( y - 1 ) * exp;
				Point pos( x, y );
				bool eastWall = IsWall( pos, DIR_E );
				bool northWall = IsWall( pos, DIR_N );
				for ( int x2 = 0; x2 < exp; x2++ )
				{
					for ( int y2 = 0; y2 < exp; y2++ )
					{
						int val = T_CORRIDOR;
						if ( m_Board[ x ][ y ] & CF_FIXED )
						{
							val = T_FLOOR;
							// Exceptions at doorways
							{
								if ( ( ! ( m_Board[ x + 1 ][ y ] & CF_FIXED ) ) && ( x2 >= exp - wall ) )
								{
									val = T_CORRIDOR;
								}
								if ( ( ! ( m_Board[ x ][ y + 1 ] & CF_FIXED ) ) && ( y2 >= exp - wall ) )
								{
									val = T_CORRIDOR;
								}
							}
						}
						if ( m_Board[ x ][ y ] & CF_KILLED ) val = T_KILLED;
						if ( eastWall && ( x2 >= exp - wall ) ) val = T_WALL;
						if ( northWall && ( y2 >= exp - wall ) ) val = T_WALL;
						m_Exp[ xP + x2 ][ yP + y2 ] = val;
					}
				}
				// Fix up odd corners
				{
					if ( eastWall )
					{
						for ( int x3 = xP + exp - wall; x3 < xP + exp; x3++ )
						{
							for ( int y3 = yP - wall; y3 < yP; y3++ )
							{
								m_Exp[ x3 ][ y3 ] = T_WALL;
							}
						}
					}
					if ( northWall )
					{
						for ( int x3 = xP - wall; x3 < xP; x3++ )
						{
							for ( int y3 = yP + exp - wall; y3 < yP + exp; y3++ )
							{
								m_Exp[ x3 ][ y3 ] = T_WALL;
							}
						}
					}
				}
			}
		}
	}
}


// Return the terrain value
int Maze::GetTerr( int x, int y ) const
{
	return m_Exp[ x ][ y ];
}

// Check whether maze is connected (changes m_Wave)
bool Maze::IsConnected()
{
	Point startPos( 1, 1 );
	Point dummy;
	int sqCount = MakeWave( startPos, MWM_COUNT, dummy );
	return ( sqCount == m_Size.x * m_Size.y );
}


 // Add walls at random, so long as maze stays connected
// This is probably the bit to improve if dungeon generation is slow
void Maze::FillWithMaze()
{
	int nWallTries = m_Size.x * m_Size.y * 2;
	for ( int wallTries = 0; wallTries < nWallTries; wallTries++ )
	{
		Point pos;
		int dir;
		int cellTries;
		for ( cellTries = 0; cellTries < 50; cellTries++ )
		{
			pos.x = 1 + random() % m_Size.x;
			pos.y = 1 + random() % m_Size.y;
			if ( m_Board[ pos.x ][ pos.y ] & CF_FIXED ) continue;
			dir = ( random() % 100 < 50 ) ? DIR_E : DIR_N;
			if ( ! IsWall( pos, dir ) ) break; 
		}
		if ( cellTries == 50 ) continue;
		Point adj = GetAdj( pos, dir );
		if ( m_Board[ adj.x ][ adj.y ] & CF_FIXED ) continue;
		MakeWall( pos, dir, true );
		if ( ! IsConnected() )
		{
			MakeWall( pos, dir, false );
		}
	}
}

// Add some rooms, which are marked with CF_FIXED so cannot subsequently get walls
void Maze::AddRooms()
{
	// Number of rooms
	int nRoomsMax = m_Params.GetNumberOfRooms();

	// Room sizes
	Point minDim;
	Point maxDim;
	m_Params.GetUnexpandedRoomSizes( minDim, maxDim );
	Point delDim = maxDim - minDim + Point( 1, 1 );

	for ( int iRoom = 0; iRoom < nRoomsMax; iRoom++ )
	{
		for ( int roomTries = 0; roomTries < 1000; roomTries++ )
		{
			Point size( minDim.x + random() % delDim.x, minDim.y + random() % delDim.y );
			Point pos( 2 + random() % ( m_Size.x - 2 - size.x ), 2 + random() % ( m_Size.y - 2 - size.y ) );

			// Rooms may not overlap (should still work if you let them, though)
			bool okay  = true;
			{
				for ( int x = pos.x - 1; x < pos.x + size.x + 1; x++ )
				{
					for ( int y = pos.y - 1; y < pos.y + size.y + 1; y++ )
					{
						if ( m_Board[ x ][ y ] & CF_FIXED )
						{
							okay = false;
							break;
						}
					}
					if ( ! okay ) break;
				}
			}
			if ( ! okay ) continue;

			MakeRoom( pos, size );
			break;			
		}
	}
}

// Create a room with one or more entrances
// If a dimension is less than 3, doors may appear in the corners
void Maze::MakeRoom( const Point & pos, const Point & size )
{
	// Rooms may not have zero size
	if ( size.x < 1 || size.y < 1 ) return;

	// Decide which sides have a door (at least one will)
	vector< bool > isEntrance;
	{
		isEntrance.push_back( true );
		for ( int iSide = 0; iSide < 4 - 1; iSide++ )
		{
			isEntrance.push_back( ( random() % 100 < 50 ) ? true : false );
		}
		pointer_to_unary_function< int, int > rg( Maze::rsRand );
		random_shuffle( isEntrance.begin(), isEntrance.end(), rg );
	}

	// Position doors randomly along each side that has one (-1 means no door)
	vector< int > places;
	{
		places.resize( 4, -1 );
		for ( int iSide = 0; iSide < 4; iSide++ )
		{
			if ( isEntrance[ iSide ] )
			{
				if ( iSide & 1 )
				{
					if ( size.y < 3 )
					{
						places[ iSide ] = 0 + random() % size.y;
					}
					else
					{
						places[ iSide ] = 1 + random() % ( size.y - 2 );
					}
				}
				else
				{
					if ( size.x < 3 )
					{
						places[ iSide ] = 0 + random() % size.x;
					}
					else
					{
						places[ iSide ] = 1 + random() % ( size.x - 2 );
					}
				}
			}
		}
	}

	for ( int x = 0; x < size.x; x++ )
	{
		for ( int y = 0; y < size.y; y++ )
		{
			m_Board[ pos.x + x ][ pos.y + y ] |= CF_FIXED;

			if ( x == 0 ) MakeWall( pos + Point( x, y ), DIR_W, y != places[ DIR_W ] );
			if ( x == size.x - 1 ) MakeWall( pos + Point( x, y ), DIR_E, y != places[ DIR_E ] );
			if ( y == 0 ) MakeWall( pos + Point( x, y ), DIR_S, x != places[ DIR_S ] );
			if ( y == size.y - 1 ) MakeWall( pos + Point( x, y ), DIR_N, x != places[ DIR_N ] );
		}
	}
}


// Add some loops by breaking walls at dead ends
void Maze::AddLoops()
{
	const int maxLoops = m_Params.GetNumberOfLoops();

	vector< Point > deadEnds;
	{
		// For simplicity, just avoid edges
		for ( int x = 2; x <= m_Size.x - 1; x++ )
		{
			for ( int y = 2; y <= m_Size.y - 1; y++ )
			{
				Point pos( x, y );
				if ( IsDeadEnd( pos ) )
				{ 
					deadEnds.push_back( pos );
				}
			}
		}
	}

	int nLoops = maxLoops;		
	if ( nLoops > deadEnds.size() ) nLoops = deadEnds.size();

	pointer_to_unary_function< int, int > rg( Maze::rsRand );
	random_shuffle( deadEnds.begin(), deadEnds.end(), rg );

	for ( int iLoop = 0; iLoop < nLoops; iLoop++ )
	{
		const Point & pos = deadEnds[ iLoop ];
		Point dummy;
		MakeWave( pos, MWM_COUNT, dummy );
		int bestVal = -1;
		int bestDir = 0;
		int iDir;
		for ( iDir = 0; iDir < 4; iDir++ )
		{
			Point adj = pos + m_Deltas[ iDir ];
			int val = m_Wave[ adj.x ][ adj.y ];
			if ( val > bestVal )
			{
				bestVal = val;
				bestDir = iDir;
			}
		}
		MakeWall( pos, bestDir, false );
	}
}


// Remove dead ends
// Each pass will create some dead ends of its own
void Maze::RemoveDeadEnds( int nPasses )
{
	for ( int pass = 0; pass < nPasses; pass++ )
	{
		// Find all the dead ends
		vector< vector< Point > > deadEnds;
		for ( int x = 1; x <= m_Size.x; x++ )
		{
			for ( int y = 1; y <= m_Size.y; y++ )
			{
				Point pos( x, y );
				if ( IsDeadEnd( pos ) )
				{
					Point fork;
					int count = MakeWave( pos, MWM_DEADEND, fork );
					vector< Point > path;
					FindPath( fork, path);
					if ( path.size() > 1 )
					{
						path.erase( path.begin() );
						deadEnds.push_back( path );
					}				
				}
			}
		}	
	
		// Remove them
		for ( int iDe = 0; iDe < deadEnds.size(); iDe++ )
		{
			const vector< Point > & de = deadEnds[ iDe ];
			for ( int iSq = 0; iSq < de.size(); iSq++ )
			{
				const Point & pos = de[ iSq ];
				m_Board[ pos.x ][ pos.y ] |= CF_KILLED;
				for ( int iDir = 0; iDir < 4; iDir++ )
				{
					MakeWall( pos, iDir , true );
				}
			}
		}
	}
}


// Is this square a dead end?
bool Maze::IsDeadEnd( const Point & pos ) const
{
	int allFlags = CF_N + CF_E + CF_S + CF_W;
	int flags = m_Board[ pos.x ][ pos.y ] & allFlags;
	return ( flags == CF_N || flags == CF_E || flags == CF_S || flags == CF_W );
}

