

// Maze generation class for roguelike maze generator
// Copyright Gerry Quinn
// 25th August, 2007


#if !defined(AFX_MAZE_H__A0C6DD2C_A882_4505_A7B4_2C44506EA773__INCLUDED_)
#define AFX_MAZE_H__A0C6DD2C_A882_4505_A7B4_2C44506EA773__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#ifdef _MSC_VER

	// MSVC - Disable warnings from long names in template expansions
	#pragma warning( disable: 4786 )

#endif


#include "Point.h"


#include <vector>
using std::vector;


class MTRand;


#define MAZE_XMAX	200						// max size of final expanded maze (changed from  original version)
#define MAZE_YMAX	200


#define EXP_MAXMUL	  6						// maximum expansion allowed (sum of wall and corridor width)


// Roguelike maze generator
// A slightly unusual approach that gives mazes of rooms and corridors slightly resembling those of classic Rogue
// I have hard-coded it in terms of number of rooms and other characteristics
// Much scope for variation, and can be adapted to make other kinds of mazes
// In a nutshell:
//		Make an empty rectangular arena of cells with 4-way connections
//		Make rooms, each with at least one door - AddRooms() 
//		Everything outside the rooms is converted into a maze of corridors - done in Create()
//		Some extra connections between corridors are created - AddLoops()
//		Most dead ends are removed - RemoveDeadEnds()
//		Maze is expanded into a classic roguelike dungeon with floor and wall tiles - Expand()
class Maze  
{
public:

	// Parameters for the maze
	class MazeParams
	{
	public:
		Point m_Size;						// size of final expanded maze - not the same as Maze::m_Size!
		int m_CorridorWidth;				// width of a corridor
		int m_WallWidth;					// width of a wall
		int m_nRooms;						// number of rooms required
		Point m_RoomMin;					// smallest room possible
		Point m_RoomMax;					// largest room possible
		int m_nLoops;						// number of loops required
		int m_nDEPasses;					// number of deadend removal passes	

		// Recommended limits, etc.			
		// Point m_Size;					max 200 x 200, 
		//									min depends on wall and corridor widths, may be as high as 25				
		// int m_CorridorWidth;				must be 1..5			
		// int m_WallWidth;					must be 1..5   
		//									total of m_WallWidth and m_Corridorwidth must be 2..6
		// int m_nRooms;									
		// Point m_RoomMin;					make it a few times (wall width + corridor width)
		// Point m_RoomMax;				
		// int m_nLoops;					your choice, about 6 often seems nice
		// int m_nDEPasses;					1 or 2 in most cases, maybe more in big mazes
		//									use a large number if you want to eliminate dead ends altogether
		//									0 will give you a different sort of dungeon
	
	public:
		int GetNumberOfLoops() const;
		int GetNumberOfRooms() const;
		int GetWallWidth() const;
		int GetCorridorWidth() const;
		Point GetSize() const;
		int GetDeadEndPasses() const;
		void GetUnexpandedRoomSizes( Point & minDim, Point & maxDim ) const;
		void GetActualRoomSizes( Point & minRoom, Point & maxRoom ) const;
		int GetExpansionFactor() const;
		MazeParams();
	};

public:

	enum Terr {
			T_KILLED = -2,					
			T_WALL = -1,
			T_FLOOR = 0,
			T_CORRIDOR
			};

protected:

	// Flags for connectivity data
	enum ConnFlags {
			CF_N = 1,						// bits for connection flag
			CF_E = 2,			
			CF_S = 4,
			CF_W = 8,
			CF_FIXED = 16,					// marks a cell whose connections may not be altered
			CF_KILLED = 32					// marks a cell that has been eliminated from the maze
			};

	// Directions (cannot be changed) 
	// Geometry note: +Y axis is north, as in standard coordinate geometry, but not most screen maps...
	enum Dir {
			DIR_N = 0,
			DIR_E,
			DIR_S,
			DIR_W
			};

	// Mode for MakeWave()
	enum MWMode {
			MWM_COUNT,						// fill all and count squares found
			MWM_DEADEND						// fill from a dead-end to a fork
			};
	
public:

	static MTRand m_Rand;												// Mersenne Twister

protected:

	Point m_Size;														// size of (unexpanded) maze

	int m_Board[ MAZE_XMAX / 2 + 1 ][ MAZE_YMAX / 2 + 1 ];				// the basic maze

	int m_Wave[ MAZE_XMAX / 2 + 1 ][ MAZE_YMAX / 2 + 1 ];				// workspace for connectivity and distance calculations

	int m_Exp[ MAZE_XMAX ][ MAZE_YMAX ];								// the expanded maze (walls and floor)

	Point m_Deltas[ 4 ];												// positions of adjacent squares

	int m_DirFlags[4];													// direction flag values

	MazeParams m_Params;												// parameters used

public:

	// Interface
	void SetParams( const MazeParams & requestedParams );				// set requested parameters
	const MazeParams & GetParams() const { return m_Params; }			// get current parameters
	void Create();														// create a maze
	int GetTerr( int x, int y ) const;									// returns a value from Maze::Terr

	Maze();
	virtual ~Maze();

	// Random functions use m_Rand
	static int random();												// general random int
	static int rsRand( int n );											// used by std::random_shuffle

protected:

	// Process maze - called by Create()
	void AddRooms();
	void MakeRoom( const Point & pos, const Point & size );
	void FillWithMaze();
	void AddLoops();
	void RemoveDeadEnds( int nPasses );
	void Expand();

	// Various calculations
	int MakeWave( const Point & startPos, MWMode mode, Point & deadendFork  );
	void FindPath( const Point & startPos, vector< Point > & path ) const;
	bool IsConnected();
	bool IsDeadEnd( const Point & pos ) const;
	
	// Simple functions
	inline Point GetAdj( const Point & cell, int dir ) const { return cell + m_Deltas[ dir ]; }
	inline bool IsWall( const Point & pos, int dir ) const { return ( ( m_Board[ pos.x ][ pos.y ] & m_DirFlags[ dir ] ) == 0 ); }

	// Make a wall (or path if 'make' is false)
	void MakeWall( const Point & pos, int dir, bool make );
};

#endif // !defined(AFX_MAZE_H__A0C6DD2C_A882_4505_A7B4_2C44506EA773__INCLUDED_)
